#!/usr/bin/perl -w

use strict;
use Data::Dumper;
use Getopt::Std;
use Storable qw(dclone);
use GD;

use vars qw ($opt_h $opt_i $opt_d $opt_s $opt_o $opt_l $opt_n $opt_D);
&getopts('hi:d:s:o:l:n:D');

my $usage = <<_EOH_;

## Options ###########################################
## Required:
# -i   a directory for organism A (used as query)
# -d   a directory for organism B (used as database)
# -s   a file containing osfinder result (*.os.txt)

## Optional:
# -l   length threshold for displaying chromosomes or contigs (default = 0)
# -o   a PNG file for output (default = <osfinder-result-file>.synteny_map.png)
# -n   number of chromosomes drawn in a line (default = 15)
# -D   set display mode (default = false)

## Others:
# -h   print help

_EOH_
    ;

# Get input parameters from command line
die $usage if $opt_h;
my $q_dir   = $opt_i or die $usage;
my $d_dir   = $opt_d or die $usage;
my $os_file = $opt_s or die $usage;
my $pngFile = $os_file . '.synteny_map.png';
$pngFile    = $opt_o if defined $opt_o;
my $cut_len = 0;
$cut_len    = $opt_l if defined $opt_l;
my $num_chr = 15;
$num_chr    = $opt_n if defined $opt_n;
my $display = 0;
$display    = 1 if $opt_D;
$q_dir     .= '/' unless $q_dir =~ /\/$/;
$d_dir     .= '/' unless $d_dir =~ /\/$/;

# Set file names
my $chrFile_1 = $q_dir . 'chrom_map';
my $chrFile_2 = $d_dir . 'chrom_map';
my @filePath1 = split('/', $q_dir);
my $orgName_1 = $filePath1[-1];
my @filePath2 = split('/', $d_dir);
my $orgName_2 = $filePath2[-1];
my (@stat_1, @stat_2);
push(@stat_1, $orgName_1);
push(@stat_2, $orgName_2);
print STDERR $orgName_1 . "\t" . $orgName_2 . "\n";

# Get the length of each chromosome
my %chr2len_1 = get_chr_length( $chrFile_1, $cut_len );
my %chr2len_2 = get_chr_length( $chrFile_2, $cut_len );

# Get the offset of each chromosome
my %chr2offset_1 = get_chr_offset( $chrFile_1, $cut_len );
my %chr2offset_2 = get_chr_offset( $chrFile_2, $cut_len );

# Get the name of each chromosome
my %chr2name_1 = get_chr_name( $chrFile_1, $cut_len );
my %chr2name_2 = get_chr_name( $chrFile_2, $cut_len );
my $num_chr2   = keys %chr2name_2;

# Read the output of OSfinder
my @os_list = read_osfinder_result( $os_file, \%chr2offset_1, \%chr2offset_2 );
my %n_coord2osinfo = cal_normalized_coordinate( \@os_list, \%chr2len_2 );

# Calculate the number of rows
my $numChr_1 = keys %chr2len_1;
my $numRow_1 = roundup( $numChr_1, $num_chr );
my $numChr_2 = keys %chr2len_2;
my $numRow_2 = roundup( $numChr_2, $num_chr );
my $maxRow   = max( $numRow_1, $numRow_2 );

# Calculate the maximum length of chromosomes
my $chrom_len_1 = max( values %chr2len_1 );
my $chrom_len_2 = max( values %chr2len_2 );
my $max_chrom_len = max( $chrom_len_1, $chrom_len_2 );

# Calculate the maximum length of chromosome names
my $name_len_1 = maximum_name_length( \%chr2name_1 );
my $name_len_2 = maximum_name_length( \%chr2name_2 );
my $max_name_len = max( $name_len_1, $name_len_2 );

# Set parameters
my $small_font_size    = 7;
my $large_font_height  = 20;

my $map_top_margin     = 10;
my $map_bottom_margin  = 10;
my $map_left_margin    = 10;
my $map_right_margin   = 10;

my $org_top_margin     = 10;
my $org_bottom_margin  = 10;
my $org_left_margin    = 10;
my $org_right_margin   = 10;
my $org_x_margin       = 50;
my $org_y_margin       = 50;

my $stat_top_margin    = 10;
my $stat_bottom_margin = 10;
my $stat_left_margin   = 10;
my $stat_right_margin  = 10;
my $stat_lines         = 1;

my $name_top_margin    = 10;
my $name_bottom_margin = 10;
my $name_height        = $max_name_len*$small_font_size;

my $chrom_width        = 20;
my $strand_width       = $chrom_width/2;
my $chrom_height       = 200;
my $chrom_x_margin     = 10;
my $chrom_y_margin     = $name_top_margin+$name_height+$name_bottom_margin;

# Calculate margin
my $org_width   = $org_left_margin+$chrom_width
    +($num_chr)*($chrom_x_margin+$chrom_width)+$org_right_margin;
my $org_height  = $org_top_margin
    +$maxRow*($chrom_height+$chrom_y_margin)+$org_bottom_margin;
my $stat_width  = $org_width;
my $stat_height = $stat_lines*$large_font_height;
my $map_width  = $map_left_margin+2*$org_width+$org_x_margin+$map_right_margin;
my $map_height = $map_top_margin+$org_height+$stat_height+$map_bottom_margin;

# Allocate map region
my @map_region = (0, 0, $map_width, $map_height);

# Allocate region of organism 1
my @org1_region = ($map_region[0]+$map_left_margin, 
		   $map_region[1]+$map_top_margin, 
		   $map_region[0]+$map_left_margin+$org_width, 
		   $map_region[1]+$map_top_margin+$org_height);

# Allocate region of organism 2
my @org2_region = ($org1_region[2]+$org_x_margin, 
		   $map_region[1]+$org_top_margin, 
		   $org1_region[2]+$org_x_margin+$org_width, 
		   $map_region[1]+$org_top_margin+$org_height);

# Allocate region of statistics 1
my @stat1_region = ($org1_region[0], 
		    $org1_region[3], 
		    $org1_region[2],
		    $org1_region[3]+$stat_height);

# Allocate region of statistics 2
my @stat2_region = ($org2_region[0], 
		    $org2_region[3], 
		    $org2_region[2], 
		    $org2_region[3]+$stat_height);

# Set up synteny map
my $syn_map    = new GD::Image($map_width, $map_height);
my $white      = $syn_map->colorAllocate(255, 255, 255);
my $black      = $syn_map->colorAllocate(0  , 0  , 0  );
$syn_map->transparent($white);
$syn_map->interlaced('true');
$syn_map->rectangle(@org1_region, $black);
$syn_map->rectangle(@org2_region, $black);

# Set the region of chromosomes
my %chr2region_1 = set_chrom_region( \@org1_region, \%chr2len_1 );
my %chr2region_2 = set_chrom_region( \@org2_region, \%chr2len_2 );
# Set the region of chromosome names
my %name2point_1 = set_name_point( \%chr2region_1, \%chr2name_1 );
my %name2point_2 = set_name_point( \%chr2region_2, \%chr2name_2 );
# Set the region of orthologous segments
my %n_coord2region_1 = set_os_region( \%n_coord2osinfo, \%chr2region_1, 0 );
my %n_coord2region_2 = set_os_region( \%n_coord2osinfo, \%chr2region_2, 1 );
my %chr2color        = cal_chr_color( \%chr2region_2 );
my %n_coord2color    = set_os_color( \%n_coord2osinfo, \%chr2color );
# Set the region of statistics
my %stat2point_1 = set_stat_point( \@stat1_region, \@stat_1 );
my %stat2point_2 = set_stat_point( \@stat2_region, \@stat_2 );

# Draw orthologous segments
draw_os( $syn_map, \%n_coord2region_1, \%n_coord2color );
draw_os( $syn_map, \%n_coord2region_2, \%n_coord2color );
# Draw chromosomes
draw_chromosome( $syn_map, \%chr2region_1 );
draw_chromosome( $syn_map, \%chr2region_2 );
# Draw chromosome names
draw_name( $syn_map, \%name2point_1 );
draw_name( $syn_map, \%name2point_2 );
# Draw statistics
draw_stat( $syn_map, \%stat2point_1 );
draw_stat( $syn_map, \%stat2point_2 );

if( $display ){
    open(DISPLAY, "| display -") or die "Can't set the display mode";
    binmode DISPLAY;
    print DISPLAY $syn_map->png;
    close DISPLAY;
} else {
    open(PNG, ">$pngFile") or die "Can't open $pngFile to write.";
    binmode PNG;
    print PNG $syn_map->png;
    close PNG;
}

sub get_chr_length {
    my $file    = shift;
    my $cut_len = shift;
    my %chr2len;
    open(IN, $file) or die "Can't open $file to read.";
    while(<IN>){
	chomp;
	my @cols = split(" ", $_);
	if( @cols >= 4 ){
	    my ($id, $chr, $len, $offset) = @cols[0, 1, 2, 3];
	    $chr2len{$id} = $len if $len>=$cut_len;
	}
    }
    close IN;
    return %chr2len;
}


sub get_chr_offset {
    my $file    = shift;
    my $cut_len = shift;
    my %chr2offset;
    open(IN, $file) or die "Can't open $file to read.";
    while(<IN>){
	chomp;
	my @cols = split(" ", $_);
	if( @cols >= 4 ){
	    my ($id, $chr, $len, $offset) = @cols[0, 1, 2, 3];
	    $chr2offset{$id} = $offset if $len>=$cut_len;
	}
    }
    close IN;
    return %chr2offset;
}


sub get_chr_name {
    my $file    = shift;
    my $cut_len = shift;
    my %chr2name;
    open(IN, $file) or die "Can't open $file to read.";
    while(<IN>){
	chomp;
	my @cols = split(" ", $_);
	if( @cols >= 4 ){
	    my ($id, $chr, $len, $offset) = @cols[0, 1, 2, 3];
	    $chr2name{$id} = $chr if $len>=$cut_len;
	}
    }
    close IN;
    return %chr2name;
}


sub read_osfinder_result {
    my $file         = shift;
    my $chr2offset_1 = shift;
    my $chr2offset_2 = shift;
    my @os_list;
    open(IN, $file) or die "Can't open $file to read.";
    while(<IN>){
	chomp;
	my @cols = split(" ", $_);
	if(@cols==8){
	    my ($chr1, $stt1, $end1, $sgn1, $chr2, $stt2, $end2, $sgn2)=@cols;
	    next if !defined( $$chr2offset_1{$chr1} );
	    next if !defined( $$chr2offset_2{$chr2} );
	    my $offset_1 = $$chr2offset_1{$chr1};
	    my $offset_2 = $$chr2offset_2{$chr2};
	    my @os_info  = ($chr1, $stt1-$offset_1, $end1-$offset_1, $sgn1, 
			    $chr2, $stt2-$offset_2, $end2-$offset_2, $sgn2);
	    push(@os_list, \@os_info);
	}
    }
    return @os_list;
}


sub cal_coordinate_offset {
    my $chr2len = shift;
    my $offset  = 0;
    my %chr2coordOffset;
    foreach my $chr (sort{$a<=>$b} keys %$chr2len){
	$chr2coordOffset{$chr} = $offset;
	$offset += $$chr2len{$chr};
    }
    return %chr2coordOffset;
}


sub cal_genome_length {
    my $chr2len = shift;
    my $genome_length = 0;
    foreach my $chr (sort{$a<=>$b} keys %$chr2len){
	$genome_length += $$chr2len{$chr};
    }
    return $genome_length;
}


sub cal_normalized_coordinate {
    my $os_list       = shift;
    my $chr2len       = shift;
    my %chr2c_offset  = cal_coordinate_offset( $chr2len );
    my $genome_length = cal_genome_length( $chr2len );

    my %n_coord2info;
    for( my $i=0 ; $i<@$os_list ; ++$i ){
	my $info    = $$os_list[$i];
	my ($chr1, $stt1, $end1, $sgn1, $chr2, $stt2, $end2, $sgn2)=@$info;
	my $coord   = $chr2c_offset{$chr2} + $stt2;
	my $n_coord = $coord / $genome_length;
	$n_coord2info{$n_coord} = $info;
    }
    return %n_coord2info;
}


sub roundup {
    my $val1 = shift;
    my $val2 = shift;
    my $flg  = $val1 % $val2;
    my $ans  = $val1 / $val2;
    $ans     =~ s/\.\d+$//;
    $ans    += 1 if $flg;
    return $ans;
}


sub max {
    my @vals = @_;
    my $max = $vals[0];
    for( my $i=1 ; $i<@vals ; ++$i ){
	$max = $vals[$i] if $vals[$i]>$max;
    }
    return $max;
}


sub maximum_name_length {
    my $chr2name = shift;
    my @names    = values %$chr2name;
    my @name_len;
    foreach (@names) { push(@name_len, length($_)); }
    return max( @name_len );
}


sub set_chrom_region {
    my $org_region = shift;
    my $chr2len    = shift;
    my %chr2region;

    my $m_left = $$org_region[0]+$org_left_margin;
    my $m_top  = $$org_region[1]+$org_top_margin;
    my $col_idx = 0;
    my $row_idx = 0;

    foreach my $chr (sort{$a<=>$b} keys %$chr2len){
	my $x1 = $m_left+$col_idx*($chrom_width+$chrom_x_margin);
	my $x2 = $x1 + $chrom_width;
	my $y1 = $m_top+$row_idx*($chrom_height+$chrom_y_margin);
	my $y2 = $y1 + $chrom_height;
	$y1    = $y2 - $chrom_height*$$chr2len{$chr}/$max_chrom_len;
	my @chrom_region = ($x1, $y1, $x2, $y2);
	$chr2region{$chr} = \@chrom_region;
	++$col_idx;
	if($col_idx==$num_chr){ ++$row_idx; $col_idx=0; }
    }
    return %chr2region;
}


sub draw_chromosome {
    my $im         = shift;
    my $chr2region = shift;
    foreach my $chr (sort{$a<=>$b} keys %$chr2region){
	my $chrom_region = $$chr2region{$chr};
	# Draw rectangle which represents the forward strand.
	my $x1 = $$chrom_region[0];
	my $x2 = $x1 + $strand_width;
	my $y1 = $$chrom_region[1];
	my $y2 = $$chrom_region[3];
	$im->rectangle($x1, $y1, $x2, $y2, $black);
	# Draw rectangle which represents the reverse strand.
	$x1        = $x2;
	$x2        = $x1 + $strand_width;
	$im->rectangle($x1, $y1, $x2, $y2, $black);
    }
}


sub set_name_point {
    my $chr2region = shift;
    my $chr2name   = shift;
    my %name2point;
    foreach my $chr (sort{$a<=>$b} keys %$chr2region){
	my $region = $$chr2region{$chr};
	my $name   = $$chr2name{$chr};
	my $x      = $$region[0];
	my $y      = $$region[3] + $name_top_margin + $name_height;
	my @point  = ($x, $y);
	$name2point{$name} = \@point;
    }
    return %name2point;
}


sub draw_name {
    my $im         = shift;
    my $name2point = shift;
    foreach my $name (sort keys %$name2point){
	my $point = $$name2point{$name};
	my $x = $$point[0];
	my $y = $$point[1];
	$im->stringUp(gdSmallFont, $x, $y, $name, $black);
    }
}


sub set_os_region {
    my $n_coord2info = shift;
    my $chr2region   = shift;
    my $org_flg      = shift;
    my %n_coord2region;
    foreach my $n_coord (sort{$a<=>$b} keys %$n_coord2info){
	my $info = $$n_coord2info{$n_coord};
	my ($chr, $stt, $end, $sgn);
	if( $org_flg ){ ($chr, $stt, $end, $sgn) = @$info[4, 5, 6, 7]; }
	else          { ($chr, $stt, $end, $sgn) = @$info[0, 1, 2, 3]; }
	my $chrom_region = $$chr2region{$chr};
	my $chrom_x1     = $$chrom_region[0];
	my $chrom_y1     = $$chrom_region[1];
	my $chrom_x2     = $$chrom_region[2];
	my $chrom_y2     = $$chrom_region[3];
	my $x1           = $chrom_x1;
	$x1             += $strand_width if $sgn eq '-';
	my $x2           = $x1 + $strand_width;
	my $y1           = $chrom_y2 - ($chrom_height*$end/$max_chrom_len);
	my $y2           = $chrom_y2 - ($chrom_height*$stt/$max_chrom_len);
	my @region       = ($x1, $y1, $x2, $y2);
	$n_coord2region{$n_coord} = \@region;
    }
    return %n_coord2region;
}


sub cal_chr_color {
    my $chr2region = shift;
    my $numChr     = keys %$chr2region;
    my %chr2color;
    my $i          = 0;
    foreach my $chr (sort{$a<=>$b} keys %$chr2region){
	my $n_coord = $i / $numChr;
	my @color   = cal_color( $n_coord );
	$chr2color{$chr} = \@color;
	++$i;
    }
    return %chr2color;
}


sub set_os_color {
    my $n_coord2info = shift;
    my $chr2color    = shift;
    my %n_coord2color;
    foreach my $n_coord (sort{$a<=>$b} keys %$n_coord2info){
	my $info  = $$n_coord2info{$n_coord};
	my $chr2  = $$info[4];
	my $color = $$chr2color{$chr2};
	$n_coord2color{$n_coord} = $color;
    }
    return %n_coord2color;
}


sub cal_color {
    my $n_coord = shift;
    my @color;
    if($n_coord < 0.2){
	my ($r, $g, $b) = (255, 0, 0);
	$g             += 128 * ($n_coord-0.0) / 0.2;
	@color          = ($r, $g, $b);
    } elsif($n_coord < 0.4) {
	my ($r, $g, $b) = (255, 128, 0);
	$g             += 127 * ($n_coord-0.2) / 0.2;
	@color          = ($r, $g, $b);
    } elsif($n_coord < 0.6) {
	my ($r, $g, $b) = (255, 255, 0);
	$r             -= 255 * ($n_coord-0.4) / 0.2;
	@color          = ($r, $g, $b);
    } elsif($n_coord < 0.8) {
	my ($r, $g, $b) = (0, 255, 0);
	$b             += 255 * ($n_coord-0.6) / 0.2;
	@color          = ($r, $g, $b);
    } else {
	my ($r, $g, $b) = (0, 255, 255);
	$g             -= 255 * ($n_coord-0.8) / 0.2;
	@color          = ($r, $g, $b);
    }
    return @color;
}


sub draw_os {
    my $im             = shift;
    my $n_coord2region = shift;
    my $n_coord2color  = shift;
    foreach my $n_coord (sort{$a<=>$b} keys %$n_coord2region){
	my $region   = $$n_coord2region{$n_coord};
	my $color    = $$n_coord2color{$n_coord};
	my $os_color;
	if( $im->colorExact(@$color) == -1 ){
	    $os_color = $im->colorAllocate(@$color);
	} else {
	    $os_color = $im->colorExact(@$color);
	}
	$im->filledRectangle(@$region, $os_color);
    }
}


sub set_stat_point {
    my $region = shift;
    my $stat   = shift;
    my $m_left = $$region[0] + $stat_left_margin;
    my $m_top  = $$region[1] + $stat_top_margin;
    my %stat2point;
    my $line_idx = 0;
    foreach my $str (@$stat){
	my $x = $m_left;
	my $y = $m_top + $line_idx*$large_font_height;
	my @point = ($x, $y);
	$stat2point{$str} = \@point;
    }
    return %stat2point;
}


sub draw_stat {
    my $im         = shift;
    my $stat2point = shift;
    foreach my $stat (sort keys %$stat2point){
	my $point   = $$stat2point{$stat};
	my ($x, $y) = @$point;
	$im->string(gdLargeFont, $x, $y, $stat, $black);
    }
}
