#!/usr/bin/perl -w

use strict;
use Data::Dumper;
use Getopt::Std;
use Storable qw(dclone);

use vars qw ($opt_h $opt_i $opt_d $opt_b $opt_o $opt_e);
&getopts('hi:d:b:o:e:');

my $usage = <<_EOH_;

## Options ###########################################
## Required:
# -i   a directory for organism A (used as query)
# -d   a directory for organism B (used as database)
# -b   a file containing blast results (-m 9 format, *.blastp)
# -o   an anchor file for output (*.anchor.txt)

## Optional:
# -e   E-value threshold for detecting anchors (default = 1e-100)

## Others:
# -h   print help

_EOH_
    ;

# Get input parameters from command line
die $usage if $opt_h;
my $q_dir     = $opt_i or die $usage;
my $d_dir     = $opt_d or die $usage;
my $blsFile   = $opt_b or die $usage;
my $ancFile   = $opt_o or die $usage;
my $cut_eval  = 1e-100;
$cut_eval     = $opt_e if defined $opt_e;
$q_dir       .= '/' unless $q_dir =~ /\/$/;
$d_dir       .= '/' unless $d_dir =~ /\/$/;

# Set file names
my $posFile_1 = $q_dir . 'all_proteins.pos';
my $chrFile_1 = $q_dir . 'chrom_map';
my $posFile_2 = $d_dir . 'all_proteins.pos';
my $chrFile_2 = $d_dir . 'chrom_map';

# Generate an anchor file 
my %acc2chr_1 = read_chr_file( $chrFile_1 );
my %id2pos_1  = read_pos_file( $posFile_1, \%acc2chr_1 );
my $num_id_1  = keys %id2pos_1;
print STDERR $num_id_1 . " markers in $posFile_1" . "\n";
my %acc2chr_2 = read_chr_file( $chrFile_2 );
my %id2pos_2  = read_pos_file( $posFile_2, \%acc2chr_2 );
my $num_id_2  = keys %id2pos_2;
print STDERR $num_id_2 . " markers in $posFile_2" . "\n";
my %pair2eval = read_blast_file( $blsFile );
my $num_hits  = keys %pair2eval;
print STDERR $num_hits . " hits in $blsFile" . "\n";
my @anchors   = detect_anchors( \%pair2eval, $cut_eval );
my $num_anc   = @anchors;
print STDERR $num_anc . " anchors were detected with E-value threthold" 
    . " ($cut_eval) " . "\n";
print_anchors( \@anchors, \%id2pos_1, \%id2pos_2, $ancFile );


sub read_pos_file {
    my $file    = shift;
    my $acc2chr = shift;
    my %id2pos;
    open(IN, $file) or die "Can't open $file to read.";
    while(<IN>){
	chomp;
	my @cols = split(" ", $_);
	if( @cols == 5 ){
	    my ($id, $acc, $stt, $end, $sgn) = @cols;
	    if( !defined( $$acc2chr{$acc} ) ){
		die "Error! : $file : $acc is not valid Accession number.";
	    }
	    my $chr = $$acc2chr{$acc};
	    if( $sgn > 0 ){ $sgn = '+'; }
	    else          { $sgn = '-'; }
	    $id2pos{$id} = join(" ", ($chr, $stt, $end, $sgn));
	}
    }
    close IN;
    return %id2pos;
}

sub read_chr_file {
    my $file = shift;
    my %acc2chr;
    open(IN, $file) or die "Can't open $file to read.";
    while(<IN>){
	chomp;
	my @cols = split(" ", $_);
	if( @cols >= 2 ){
	    my ($chr, $acc) = @cols[0, 1];
	    $acc2chr{$acc}  = $chr;
	}
    }
    close IN;
    return %acc2chr;
}

sub read_blast_file {
    my $file = shift;
    my %pair2eval;
    open(IN, $file) or die "Can't open $file to read.";
    while(<IN>){
	chomp;
	next if /^#/;
	my @cols = split(" ", $_);
	if( @cols == 12 ){
	    my ($query, $subject, $e_val) = @cols[0, 1, 10];
	    my $pair = make_pair($query, $subject);
	    $pair2eval{$pair} = $e_val;
	}
    }
    close IN;
    return %pair2eval;
}


sub detect_anchors {
    my $pair2eval = shift;
    my $cut_eval  = shift;
    my @anchors;
    foreach my $pair (keys %$pair2eval){
	my $e_val = $$pair2eval{$pair};
	if( $e_val <= $cut_eval ){
	    push( @anchors, $pair );
	}
    }
    return @anchors;
}


sub print_anchors {
    my $anchors  = shift;
    my $id2pos_1 = shift;
    my $id2pos_2 = shift;
    my $ancFile  = shift;
    open(OUT, ">$ancFile") or die "Can't open $ancFile to write.";
    foreach my $anc ( @$anchors ){
	my ($id_1, $id_2) = get_pair_id( $anc );
	my $pos_1 = $$id2pos_1{$id_1};
	my $pos_2 = $$id2pos_2{$id_2};
	print OUT join(" ", ($pos_1, $pos_2)) . "\n";
    }
    close OUT;
}


sub make_pair {
    my $val1 = shift;
    my $val2 = shift;
    return $val1 . ' ' . $val2;
}


sub get_pair_id {
    my $pair = shift;
    return split(" ", $pair);
}
