#!/usr/bin/perl -w

use strict;
use Data::Dumper;
use Getopt::Std;
use Storable qw(dclone);

use vars qw ($opt_h $opt_i);
&getopts('hi:');

my $usage = <<_EOH_;

## Options ###########################################
## Required:
# -i   a directory containing sequence files downloaded from Ensembl

## Others:
# -h   print help

_EOH_
    ;

# Get input parameters from command line
die $usage if $opt_h;
my $input = $opt_i or die $usage;
$input   .= '/' unless $input =~ /\/$/;

# Set file names
my $mfaFile = $input . "all_proteins.mfa";
my $posFile = $input . "all_proteins.pos";
my $chrFile = $input . "chrom_map";
my $in_file = $input . "ensembl_mfa";

# Parse a ensembl_mfa file
my %gid2info = read_ensembl_mfa( $in_file );
my $num_gid  = keys %gid2info;
print STDERR $num_gid . " genes were found in $in_file.\n";

# Print mfa and position file
open(MFA, ">$mfaFile") or die "Can't open $mfaFile to write.";
open(POS, ">$posFile") or die "Can't open $posFile to write.";
my (%chr2min, %chr2max);
foreach my $gid ( sort keys %gid2info ){
    my $info = $gid2info{$gid};
    my $pos  = $$info[0];
    my $seq  = $$info[1];
    my ($chr, $stt, $end, $sgn) = split(" ", $pos);
    if( !defined($chr2min{$chr}) or $stt<$chr2min{$chr} ){
	$chr2min{$chr} = $stt;
    }
    if( !defined($chr2max{$chr}) or $end>$chr2max{$chr} ){
	$chr2max{$chr} = $end;
    }
    print MFA ">" . $gid . "\n" . $seq . "\n";
    print POS $gid . "\t" . $pos . "\n";
}
close MFA;
close POS;

# Print chromosome map
my %chrom_map;
my $max_chr = 0;
foreach my $chr (sort keys %chr2min){
    if( $chr =~ /^(\d+)$/ ){
	my $id = $chr;
	$chrom_map{$id} = $chr;
	$max_chr = $chr if $chr > $max_chr;
    }
}
my $chrom_id = $max_chr;
foreach my $chr (sort keys %chr2min){
    unless( $chr =~ /^(\d+)$/ ){
	$chrom_map{++$chrom_id} = $chr;
    }
}
open(CHR, ">$chrFile") or die "Can't open $chrFile to write.";
foreach my $id ( sort{$a<=>$b} keys %chrom_map ){
    my $chr = $chrom_map{$id};
    my $len = $chr2max{$chr} - $chr2min{$chr};
    my $min = $chr2min{$chr};
    print CHR $id . "\t" . $chr . "\t" . $len . "\t" . $min . "\n";
}
close CHR;


sub read_ensembl_mfa {
    my $file = shift;
    open(IN, $file) or die "Can't open $file to read.";
    my ($pid, $tid, $gid, $pos, $seq);
    my (%gid2seq, %gid2pos, %gid2info);
    while(<IN>){
	chomp;
	if(/>(\S+) \w+:\w+ (\S+) gene:(\S+) transcript:(\S+)/){
	    if($seq){
		$gid2seq{$gid} = $seq;
		if( $gid2info{$gid} ){
		    my $tmp_list = $gid2info{$gid};
		    if(length($gid2seq{$gid}) > length( $$tmp_list[1] )){
			my @tmpList; 
			push( @tmpList, $gid2pos{$gid} );
			push( @tmpList, $gid2seq{$gid} );
			$gid2info{$gid} = \@tmpList;
		    }
		} else {
		    my @tmpList; 
		    push( @tmpList, $gid2pos{$gid} );
		    push( @tmpList, $gid2seq{$gid} );
			$gid2info{$gid} = \@tmpList;
		}
		undef $seq;
	    }
	    ($pid, $pos, $gid, $tid) = ($1, $2, $3, $4);
	    my @posCols = split(':', $pos);
	    my $pos     = join("\t", @posCols[2..5]);
	    $gid2pos{$gid} = $pos;
	} else {
	    $seq .= $_;
	} 
    }
    if($seq){
	$gid2seq{$gid} = $seq;
	if( $gid2info{$gid} ){
	    my $tmp_list = $gid2info{$gid};
	    if(length($gid2seq{$gid}) > length( $$tmp_list[1] )){
		my @tmpList; 
		push( @tmpList, $gid2pos{$gid} );
		push( @tmpList, $gid2seq{$gid} );
		$gid2info{$gid} = \@tmpList;
	    } else {
		my @tmpList; 
		push( @tmpList, $gid2pos{$gid} );
		push( @tmpList, $gid2seq{$gid} );
			$gid2info{$gid} = \@tmpList;
	    }
	}
	undef $seq;
    }
    close IN;
    return %gid2info;
}
