#!/usr/bin/perl -w

use strict;
use Data::Dumper;
use Getopt::Std;
use Storable qw(dclone);
use Bio::SeqIO;

use vars qw ($opt_h $opt_i);
&getopts('hi:');

my $usage = <<_EOH_;

## Options ###########################################
## Required:
# -i   a directory containing sequence files in the GBK format

## Others:
# -h   print help

_EOH_
    ;

# Get input parameters from command line
die $usage if $opt_h;
my $input = $opt_i or die $usage;
$input   .= '/' unless $input =~ /\/$/;

# Set file names
my (@gbk_files, $mfaFile, $posFile, $chrFile);
die "$input is not a valid directory." unless (-d $input);
opendir( DIR, $input ) or die "Can't open $input as a directory.";
while( my $file=readdir(DIR) ){
    if( $file =~ /^(.+)\.gbk$/ ){
	push( @gbk_files, $input.$file );
    }
}
$mfaFile = $input . "all_proteins.mfa";
$posFile = $input . "all_proteins.pos";
$chrFile = $input . "chrom_map";
die "Error! : No gbk files are included in $input.\n" . $usage 
    unless @gbk_files;

# Parse GBK files
open(MFA, ">$mfaFile") or die "Can't open $mfaFile to write.";
open(POS, ">$posFile") or die "Can't open $posFile to write.";
open(CHR, ">$chrFile") or die "Can't open $chrFile to write.";
my $chr_id = 0;
foreach my $gbkFile ( @gbk_files ){
    print STDERR "Parsing $gbkFile..." . "\n";
    my $seq_in  = Bio::SeqIO->new(-file   => $gbkFile, 
				  -format => 'genbank')
	or die "Can't open $gbkFile." . "\n"
	. "$gbkFile may not be formatted in the GBK format." . "\n";
    while( my $seq = $seq_in->next_seq() ){
	my $acc_num  = $seq->id;
	print STDERR "Accession Number   : " . $acc_num . "\n";
	my $desc     = $seq->desc;
	print STDERR "Description        : " . $desc . "\n";
	my $length   = $seq->length;
	print STDERR "Length             : " . $length . "\n";
	my $offset   = 1;
	print STDERR "Offset             : " . $offset . "\n";
	my @features = $seq->get_SeqFeatures;
	my $num_feat = @features;
	print STDERR "Number of features : " . $num_feat . "\n";
	my $num_cds  = 0;
	foreach my $feat (@features){
	    if($feat->primary_tag eq 'CDS'){
		my $id  = join(' ', $feat->each_tag_value('gene'))
		    || join(' ', $feat->each_tag_value('locus_tag'))
		    || join(' ', $feat->each_tag_value('protein_id'));
		my $stt = $feat->start;
		my $end = $feat->end;
		my $sgn = $feat->strand;
		my $seq = join(' ', $feat->each_tag_value('translation'));
		if($id and $stt and $end and $sgn and $seq){
		    print MFA ">" . $id . "\n" . $seq . "\n";
		    print POS $id . "\t" . $acc_num . "\t"
			. $stt . "\t" . $end . "\t" . $sgn . "\n";
		    ++$num_cds;
		}
	    }
	}
	print STDERR "Number of markers  : " . $num_cds . "\n\n";
	next unless $num_cds;
	print CHR ++$chr_id . "\t" . $acc_num . "\t" 
	    . $length . "\t" . $offset . "\t" . $desc . "\n";
    }
}
close MFA;
close POS;
close CHR;
