// *************************************************************************
//  learn.cpp
// *************************************************************************

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "osfinder.h"

// *************************************************************************
// initialize parameters.
void initialize(double& Ps_mg, double& Ps_gm, double& Pr_mg, double& Pr_gm, const std::vector<block>& anc_set, const std::vector<edge>& sn_edge_set, int num_org){
  double ancLenSum=0;
  double ancLenMin, ancLenMax;
  for(int i=0;i<(int)anc_set.size();++i){
    if(i==0){
      ancLenMin=anc_set[i].len;
      ancLenMax=anc_set[i].len;
    }
    ancLenSum+=anc_set[i].len;
    if(anc_set[i].len<ancLenMin) ancLenMin=anc_set[i].len;
    if(anc_set[i].len>ancLenMax) ancLenMax=anc_set[i].len;
  }
  double ancLenAve=ancLenSum/(int)anc_set.size();
  Ps_mg=2*num_org/(ancLenAve+ancLenMax);
  Pr_mg=2*num_org/(ancLenAve+ancLenMin);

  double edgeLenSum=0;
  double edgeLenMin, edgeLenMax;
  for(int i=0;i<(int)sn_edge_set.size();++i){
    if(i==0){
      edgeLenMin=sn_edge_set[i].len;
      edgeLenMax=sn_edge_set[i].len;
    }
    edgeLenSum+=sn_edge_set[i].len;
    if(sn_edge_set[i].len<edgeLenMin) edgeLenMin=sn_edge_set[i].len;
    if(sn_edge_set[i].len>edgeLenMax) edgeLenMax=sn_edge_set[i].len;
  }

  double edgeLenAve=edgeLenSum/(int)sn_edge_set.size();
  Ps_gm=2*num_org/(edgeLenAve+edgeLenMin);
  Pr_gm=2*num_org/(edgeLenAve+edgeLenMax);
}

// *************************************************************************
// learn parameters.
void learn( const std::vector<edge>& sn_edge_set, const std::vector<block>& anc_set, double& Ps_mg, double& Ps_gm, double& Pr_mg, double& Pr_gm, int num_org ){
  double Ps_mm = 1.0 - Ps_mg;
  double Ps_gg = 1.0 - Ps_gm;
  double Pr_mm = 1.0 - Pr_mg;
  double Pr_gg = 1.0 - Pr_gm;

  double s_mm_scr = log( Ps_mm );
  double s_mg_scr = log( Ps_mg );
  double s_gm_scr = log( Ps_gm );
  double s_gg_scr = log( Ps_gg );
  double r_mm_scr = log( Pr_mm );
  double r_mg_scr = log( Pr_mg );
  double r_gm_scr = log( Pr_gm );
  double r_gg_scr = log( Pr_gg );

  std::vector<int> SHR_edge_len;
  std::vector<int> NSHR_edge_len;
  double edge_scr = 0.0;
  for( int i=0 ; i<(int)sn_edge_set.size() ; ++i ){
    double SHR_scr = cal_SHR_edge_scr( sn_edge_set[i].len, s_gg_scr, s_gm_scr, num_org );
    double NSHR_scr = cal_NSHR_edge_scr( sn_edge_set[i].len, r_gg_scr, r_gm_scr, num_org );
    if( SHR_scr > NSHR_scr ){
      SHR_edge_len.push_back( sn_edge_set[i].len );
      edge_scr += SHR_scr;
    } else {
      NSHR_edge_len.push_back( sn_edge_set[i].len );
      edge_scr += NSHR_scr;
    }
  }

  Ps_gm = refine_Ps_gm( SHR_edge_len, num_org );
  if( Ps_gm < 0 ){
    exit(1);
  }
  Pr_gm = refine_Pr_gm( NSHR_edge_len, num_org );
  if( Pr_gm < 0 ){
    exit(1);
  }


  std::vector<int> SHR_anc_len;
  std::vector<int> NSHR_anc_len;
  double anc_scr = 0.0;
  for( int i=0 ; i<(int)anc_set.size() ; ++i ){
    double SHR_scr = cal_SHR_anc_scr( anc_set[i].len, s_mm_scr, s_mg_scr, num_org );
    double NSHR_scr = cal_NSHR_anc_scr( anc_set[i].len, r_mm_scr, r_mg_scr, num_org );
    if( SHR_scr > NSHR_scr ){
      SHR_anc_len.push_back( anc_set[i].len );
      anc_scr += SHR_scr;
    } else {
      NSHR_anc_len.push_back( anc_set[i].len );
      anc_scr += NSHR_scr;
    }
  }

  Ps_mg = refine_Ps_mg( SHR_anc_len, num_org );
  if( Ps_mg < 0 ){
    exit(1);
  }
  Pr_mg = refine_Pr_mg( NSHR_anc_len, num_org );
  if( Pr_mg < 0 ){
    exit(1);
  }

  // std::cout << "edge score   : " << edge_scr << std::endl;
  // std::cout << "anchor score : " << anc_scr << std::endl;
}


// *************************************************************************
// learn parameters in anchor models.
void learn_anchor_model( const std::vector<block>& anc_set, double& Ps_mg, double& Pr_mg, int num_org ){
  std::vector<int> anc_len_set;

  for( int i=0 ; i<(int)anc_set.size() ; ++i ){
    anc_len_set.push_back(anc_set[i].len);
  }

  sort( anc_len_set );
  double max_anchor_scr=0.0;
  double max_Ps_mg, max_Pr_mg;
  int max_idx;
  for( int i=0 ; i<(int)anc_len_set.size()-1 ; ++i ){
    std::vector<int> SHR_anc_len;
    std::vector<int> NSHR_anc_len;
    for( int j=0 ; j<(int)anc_len_set.size() ; ++j ){
      if(j<=i) NSHR_anc_len.push_back(anc_len_set[j]);
      else SHR_anc_len.push_back(anc_len_set[j]);
    }    
    Ps_mg = refine_Ps_mg( SHR_anc_len, num_org );
    Pr_mg = refine_Pr_mg( NSHR_anc_len, num_org );

    double Ps_mm = 1.0 - Ps_mg;
    double Pr_mm = 1.0 - Pr_mg;

    double s_mm_scr = log( Ps_mm );
    double s_mg_scr = log( Ps_mg );
    double r_mm_scr = log( Pr_mm );
    double r_mg_scr = log( Pr_mg );

    double anchor_scr=0.0;
    for( int j=0 ; j<(int)anc_len_set.size() ; ++j ){
      if(j<=i) anchor_scr+=cal_NSHR_anc_scr( anc_len_set[j], r_mm_scr, r_mg_scr, num_org );
      else anchor_scr+=cal_SHR_anc_scr( anc_len_set[j], s_mm_scr, s_mg_scr, num_org );
    }  
  
    if(max_anchor_scr==0 || anchor_scr > max_anchor_scr){
      max_anchor_scr = anchor_scr;
      max_Ps_mg = Ps_mg;
      max_Pr_mg = Pr_mg;
      max_idx = i;
    }
  }

  Ps_mg = max_Ps_mg;
  Pr_mg = max_Pr_mg;
}

// *************************************************************************
// learn parameters in edge models.
void learn_edge_model( const std::vector<edge>& sn_edge_set, double& Ps_gm, double& Pr_gm, int num_org ){
  std::vector<int> edge_len_set;

  for( int i=0 ; i<(int)sn_edge_set.size() ; ++i ){
    edge_len_set.push_back(sn_edge_set[i].len);
  }

  sort( edge_len_set );
  double max_edge_scr=0;
  double max_Ps_gm, max_Pr_gm;
  int max_idx;
  for( int i=0 ; i<(int)edge_len_set.size()-1 ; ++i ){
    std::vector<int> SHR_edge_len;
    std::vector<int> NSHR_edge_len;
    for( int j=0 ; j<(int)edge_len_set.size() ; ++j ){
      if(j<=i) SHR_edge_len.push_back(edge_len_set[j]);
      else NSHR_edge_len.push_back(edge_len_set[j]);
    }    
    Ps_gm = refine_Ps_gm( SHR_edge_len, num_org );
    Pr_gm = refine_Pr_gm( NSHR_edge_len, num_org );

    double Ps_gg = 1.0 - Ps_gm;
    double Pr_gg = 1.0 - Pr_gm;

    double s_gg_scr = log( Ps_gg );
    double s_gm_scr = log( Ps_gm );
    double r_gg_scr = log( Pr_gg );
    double r_gm_scr = log( Pr_gm );

    double edge_scr=0.0;
    for( int j=0 ; j<(int)edge_len_set.size() ; ++j ){
      if(j<=i) edge_scr+=cal_SHR_anc_scr( edge_len_set[j], s_gg_scr, s_gm_scr, num_org );
      else edge_scr+=cal_NSHR_anc_scr( edge_len_set[j], r_gg_scr, r_gm_scr, num_org );
    }  
  
    if(max_edge_scr ==0 || edge_scr > max_edge_scr){
      max_edge_scr = edge_scr;
      max_Ps_gm = Ps_gm;
      max_Pr_gm = Pr_gm;
      max_idx = i;
    }
  }

  Ps_gm = max_Ps_gm;
  Pr_gm = max_Pr_gm;
}


// *************************************************************************
// sort according to its value
void sort( std::vector<int>& vec ){
  std::map< int, int > length_cnt;

  for( int i=0 ; i<(int)vec.size() ; ++i ){
    std::map< int, int >::iterator p;
    p = length_cnt.find( vec[i] );
    if( p != length_cnt.end() ){
      (p->second)+=1;
    } else {
      length_cnt.insert( std::map<int, int>::value_type(vec[i], 1 ) );
    }
  }

  vec.clear();
  std::map< int, int >::iterator p = length_cnt.begin();
  while( p != length_cnt.end() ){
    int len = p->first;
    int cnt = p->second;
    for( int i=0 ; i<cnt ; ++i ){
      vec.push_back( len );
    }
    ++p;
  }

  /*
  for( int i=0 ; i<(int)vec.size() ; ++i ){
    int min = vec[i];
    int min_idx = i;
    for( int j=i+1 ; j<(int)vec.size() ; ++j ){
      if( vec[j] < min ){
	min = vec[j];
	min_idx = j;
      }
    }
    vec[min_idx] = vec[i];
    vec[i] = min;
  }
  */
}


// *************************************************************************
// calculate log-likelihood of an edge based on SHR model.
double cal_SHR_edge_scr( int edge_len, double s_gg_scr, double s_gm_scr, int num_org ){
  double scr = ( edge_len - num_org ) * s_gg_scr + num_org * s_gm_scr; 
  return scr;
}

// *************************************************************************
// calculate log-likelihood of an edge based on NSHR model.
double cal_NSHR_edge_scr( int edge_len, double r_gg_scr, double r_gm_scr, int num_org ){
  double scr = ( edge_len - num_org ) * r_gg_scr + num_org * r_gm_scr;
  return scr;
}

// *************************************************************************
// calculate log-likelihood of an anchor based on SHR model.
double cal_SHR_anc_scr( int anc_len, double s_mm_scr, double s_mg_scr, int num_org ){
  double scr = ( anc_len - num_org ) * s_mm_scr + num_org * s_mg_scr; 
  return scr;
}

// *************************************************************************
// calculate log-likelihood of an anchor based on NSHR model.
double cal_NSHR_anc_scr( int anc_len, double r_mm_scr, double r_mg_scr, int num_org ){
  double scr = ( anc_len - num_org ) * r_mm_scr + num_org * r_mg_scr;
  return scr;
}

// *************************************************************************
// refine Ps_gm.
double refine_Ps_gm( std::vector<int> SHR_edge_len, int num_org ){
  int freq_gm = (int)SHR_edge_len.size() * num_org;
  double freq_all = 0;
  for( int i=0 ; i<(int)SHR_edge_len.size() ; ++i ){
    freq_all += SHR_edge_len[i];
  }

  double Ps_gm = (double)freq_gm / (double)freq_all;
  //  std::cout << "Ps_gm = " << freq_gm << " / " << freq_all << "\n";
  return Ps_gm;
}

// *************************************************************************
// refine Pr_gm.
double refine_Pr_gm( std::vector<int> NSHR_edge_len, int num_org ){
  int freq_gm = (int)NSHR_edge_len.size() * num_org;
  double freq_all = 0;
  for( int i=0 ; i<(int)NSHR_edge_len.size() ; ++i ){
    freq_all += NSHR_edge_len[i];
  }

  double Pr_gm = (double)freq_gm / (double)freq_all;
  //std::cout << "Pr_gm = " << freq_gm << " / " << freq_all << "\n";
  return Pr_gm;
}

// *************************************************************************
// refine Ps_mg.
double refine_Ps_mg( std::vector<int> SHR_anc_len, int num_org ){
  int freq_mg = (int)SHR_anc_len.size() * num_org;
  double freq_all = 0;
  for( int i=0 ; i<(int)SHR_anc_len.size() ; ++i ){
    freq_all += SHR_anc_len[i];
  }

  double Ps_mg = (double)freq_mg / (double)freq_all;
  //  std::cout << "Ps_mg = " << freq_mg << " / " << freq_all << "\n";
  return Ps_mg;
}

// *************************************************************************
// refine Pr_gm.
double refine_Pr_mg( std::vector<int> NSHR_anc_len, int num_org ){
  int freq_mg = (int)NSHR_anc_len.size() * num_org;
  double freq_all = 0;
  for( int i=0 ; i<(int)NSHR_anc_len.size() ; ++i ){
    freq_all += NSHR_anc_len[i];
  }

  double Pr_mg = (double)freq_mg / (double)freq_all;
  //  std::cout << "Pr_mg = " << freq_mg << " / " << freq_all << "\n";
  return Pr_mg;
}
